package RPM::Source::Dependency::Analyzer::Storage;

use strict;
use warnings;
use Carp;

our $VERSION=0.0003;

our $verbose=0;

my @_DEPS_KEYS=qw/girs hdrs libs path perl pkgconf pkgs python ruby vapi cmake/;
my %SA_STORAGE_KEY=(
'gir','girs',
'vapi','vapi',
'devel-libs','libs',
'headers','hdrs',
'headers-rebased','hdrs',
'lua','lua',
'packages', 'pkgs',
'path','path',
'perl','perl',
'pkg-config','pkgconf',
'python2','python',
'python3','python',
'ruby','ruby',
'cmake','cmake',
);
my %_DEBUG_NAMES=qw/
libs devel-libs
hdrs headers
path executables
pkgs packages
girs gir-files
pkgconf pkgconfig
/;

my %known_lang=map{$_=>1} qw/
c
cxx
fortran
go
lua
perl_xs
perl_buildscripts
perl_inc_Module_AutoInstall
perl_inc_Module_Install
prog_pod2man
python
ruby
swig
texinfo
/;

sub new {
    my $class = shift;
    my @key_storages=map {
	$_.'_ignore' => [],
	$_.'_req' => []
    } @_DEPS_KEYS;

    my $self = {
	@key_storages,
	lang=>{},
	@_
    };
    bless $self, $class;
}

sub dump {
    my ($storage)=@_;
    my $sorted=$storage->__create_dump_arrayref();
    map {
    print $_->[0],"\t",$_->[1],"\t",$_->[2],"\n";
    } @$sorted;
}

sub dump_to_file {
    my ($storage,$filename)=@_;
    my $sorted=$storage->__create_dump_arrayref();
    open (my $fn, '>', $filename) || die "dump_to_file('$filename'): $!. Stopped";
    map {
    print $fn $_->[0],"\t",$_->[1],"\t",$_->[2],"\n";
    } @$sorted;
    close ($fn) || die "dump_to_file('$filename'): $!. Stopped";
}

sub __create_dump_arrayref {
    my ($storage)=@_;
    $storage->_uniquify_storage();
    my @text_to_sort;
    foreach my $lang ( keys(%{$storage->{lang}}) ) {
	my $val=$storage->{lang}->{$lang};
	push @text_to_sort, ['langflag',$lang, $val ? '1' : '0'];
    }
    foreach my $deproot (@_DEPS_KEYS) {
	foreach my $type ('ignore','req') {
	    my $key =$deproot.'_'.$type;
	    my $typestring=$type eq 'ignore'? $type: 'require';
	    foreach my $val (@{$storage->{$key}}) {
		if (ref $val) {
		    my $text='['.join(':',sort @$val).']';
		    push @text_to_sort, [$typestring,$deproot,$text];
		} else {
		    push @text_to_sort, [$typestring,$deproot,$val];
		}
	    }
	}
    }
    my @sorted=sort {
	my $cmp=$a->[0] cmp $b->[0];
	$cmp=$a->[1] cmp $b->[1] if $cmp==0;
	$cmp=$a->[2] cmp $b->[2] if $cmp==0;
	return $cmp;
    } @text_to_sort;
    return \@sorted;
}

sub _uniquify_storage {
    my ($storage)=@_;
    foreach (@_DEPS_KEYS) {
	foreach my $key ($_.'_ignore',$_.'_req') {
	    my @new_list=&__unique_list(@{$storage->{$key}});
	    $storage->{$key}=\@new_list;
	}
    }
}

sub __unique_list {
    # array for alternative choices
    my @out;
    my %req;
    foreach my $a (@_) {
	if (not defined $a) {
	    Carp::cluck "undef in __unique_list";
	} elsif (ref($a)) {
	    if (ref($a) eq 'ARRAY') {
		if (@$a == 0) {
		    Carp::carp "Oops! empty ref in __unique_list";
		} elsif (@$a == 1) {
		    # unarray it
		    $req{$a->[0]}=1;
		} else {
		    push @out, $a;
		}
	    } else {
		Carp::cluck "bad ref in __unique_list: $a";
	    }
	} else {
	    $req{$a}=1;
	}
    }
    return @out, keys %req;
}

### refactoring subroutine
sub get_req_arrayref_by_SA_dbname($$) {
    my ($storage,$dbname)=@_;
    my $key=$SA_STORAGE_KEY{$dbname};
    die "internal error: no storage key for $dbname" unless $key;
    return $storage->{$key.'_req'};
}

### refactoring subroutine
sub get_ignore_arrayref_by_SA_dbname($$) {
    my ($storage,$dbname)=@_;
    my $key=$SA_STORAGE_KEY{$dbname};
    die "internal error: no storage key for $dbname" unless $key;
    return $storage->{$key.'_ignore'};
}


# sub __unique_and_filter_list {
#     my %ignore=map {$_=>1} @{shift()};
#     my @out;
#     my %req;
#     foreach my $a (@_) {
# 	if (ref($a)) {
# 	    die "Oops: internal error: not an ARRAY" if ref ($a) ne 'ARRAY';
# 	    push @out, [ grep {not $ignore{$_}} @$a];
# 	} elsif (not $ignore{$a}) {
# 	    $req{$a}=1;
# 	}
#     }
#     return @out, keys %req;
# }

sub __DEBUG_NAMES {
    my $debugname=$_DEBUG_NAMES{$_[0]};
    return $debugname ? $debugname : $_[0];
}

sub _debug_storage {
    my ($storage,$inputname)=@_;
    foreach my $dbkey (@_DEPS_KEYS) {
	&__debug_msg($inputname,'BR '.&__DEBUG_NAMES($dbkey),$storage->{$dbkey.'_req'});
    }
    foreach my $dbkey (@_DEPS_KEYS) {
	&__debug_msg($inputname,'ignore '.&__DEBUG_NAMES($dbkey),$storage->{$dbkey.'_ignore'});
    }
}

sub __debug_msg {
    my ($file,$br_name,$list)=@_;
    return unless @$list;
    print STDERR "$file: $br_name(detected):",
    join(' ',map {ref($_) eq 'ARRAY' ? '[ '.join(' ',@$_).' ]' : $_} @$list),"\n";
}

sub __check_arg_ref {
	# some checks
    ...
#    foreach my $i (@_) {
#	Carp::cluck 'got an array:['.join(',',@$i).']' if ref $i;
#    }
}

sub add_XXX_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_XXX_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{XXX_req}}, @_;
}
sub add_XXX_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_XXX_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{XXX_ignore}}, @_;
}

sub add_raw_storage_deps {
    my $self=shift;
    my $class=shift;
    my $type=shift;
    $type='_ignore' if $type eq 'ignore';
    print STDERR "DEBUG:Collect: add_raw_storage_deps: $class $type ",join(',',@_),"\n" if $verbose;
    push @{$self->{$class.$type}}, @_;
}
sub get_lang_flag {
    my ($self,$lang)=@_;
    Carp::confess "unsupported lang $_[1]" unless $known_lang{$_[1]};
    print STDERR "DEBUG:Collect: add_XXX_req: ",join(',',@_),"\n" if $verbose;
    return $self->{'lang'}->{$lang};
}
sub set_lang_flag {
    my ($self,$lang)=@_;
    Carp::confess "unsupported lang $_[1]" unless $known_lang{$_[1]};
    print STDERR "DEBUG:Collect: set_lang_flag: $lang=1\n" if $verbose;
    $self->{'lang'}->{$lang}=1;
}
sub clear_lang_flag {
    my ($self,$lang)=@_;
    Carp::confess "unsupported lang $_[1]" unless $known_lang{$_[1]};
    print STDERR "DEBUG:Collect: clear_lang_flag: $lang\n" if $verbose;
    delete $self->{'lang'}->{$lang};
}

sub add_lua_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_lua_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{lua_req}}, @_;
}
sub add_lua_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_lua_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{lua_ignore}}, @_;
}
sub add_ruby_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_ruby_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{ruby_req}}, @_;
}
sub add_ruby_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_ruby_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{ruby_ignore}}, @_;
}
sub add_python_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_python_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{python_req}}, @_;
}
sub add_python_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_python_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{python_ignore}}, @_;
}
sub add_libs_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_libs_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{libs_req}}, @_;
}
sub add_libs_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_libs_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{libs_ignore}}, @_;
}
sub add_hdrs_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_hdrs_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{hdrs_req}}, @_;
}
sub add_hdrs_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_hdrs_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{hdrs_ignore}}, @_;
}
sub add_path_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_path_req: ",join(',',@_),"\n" if $verbose;
    #push @{$self->{path_req}}, @_;
    foreach my $path (@_) {
	next unless $path;
	my @out;
	my @in=($path);
	@in=@$path if ref $path;
	local $_;
	my $before;
	foreach (@in) {
	    my $before=$_;
	    s/^["']//;s/["']$//;s/\s.*$//;s/\$EXEEXT$//;
	    Carp::carp "add_path_req: bad input [$before]: mapped to [$_]\n" if $_ ne $before;
	    push @out,$_;
	}
	if (ref $path) {
	    push @{$self->{path_req}}, \@out;
	} else {
	    push @{$self->{path_req}}, @out;
	}
    }
}
sub add_path_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_path_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{path_ignore}}, @_;
}
sub add_perl_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_perl_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{perl_req}}, @_;
}
sub add_perl_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_perl_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{perl_ignore}}, @_;
}
sub add_pkgs_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_pkgs_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{pkgs_req}}, @_;
}
sub add_pkgs_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_pkgs_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{pkgs_ignore}}, @_;
}
sub add_girs_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_girs_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{girs_req}}, @_;
}
sub add_girs_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_girs_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{girs_ignore}}, @_;
}
sub add_vapi_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_vapi_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{vapi_req}}, @_;
}
sub add_vapi_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_vapi_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{vapi_ignore}}, @_;
}
sub add_pkgconf_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_pkgconf_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{pkgconf_req}}, @_;
}
sub add_pkgconf_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_pkgconf_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{pkgconf_ignore}}, @_;
}
sub add_cmake_req {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_cmake_req: ",join(',',@_),"\n" if $verbose;
    push @{$self->{cmake_req}}, @_;
}
sub add_cmake_ignore {
    my ($self)=shift;
    #&__check_arg_ref;
    print STDERR "DEBUG:Collect: add_cmake_ignore: ",join(',',@_),"\n" if $verbose;
    push @{$self->{cmake_ignore}}, @_;
}


1;
